#include "HARK_Matrix.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <iostream>

#ifdef USING_LAPACKE_MATRIX

#include "cblas.h"
#include "lapacke.h"

#else

#include "svd.hpp"
#include "evd.hpp"

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#endif

char harkmatrixtype[] = "complex";

namespace HARKCORE
{
    class IndexOutOfBoundException : public std::exception
    {
    };

    class MatrixSizeMismatchException : public std::exception
    {
    };

    class NotSquareMatrixException : public std::exception
    {
    };

    class NotVectorException : public std::exception
    {
    };

    class NotScalarException : public std::exception
    {
    };

#ifdef USING_LAPACKE_MATRIX
    class FailedLAPACKEException : public std::exception
    {
    };
#endif

#ifdef USING_HARKIO3_MATRIX
    Matrix::Matrix() : rows(0), cols(0), mat_value(NULL)
    {
    }

    Matrix::Matrix(int rows, int cols) : rows(rows), cols(cols), mat_value(NULL)
    {
		if (rows == 0 || cols == 0){
			rows = 0;
			cols = 0;
			return;
		}
		this->mat_value = harkio_Matrix_zeros(harkmatrixtype, rows, cols);
		if (this->mat_value != NULL){
			this->rows = rows;
			this->cols = cols;
		}
    }

    Matrix::Matrix(const Matrix &src) : rows(src.rows), cols(src.cols), mat_value(NULL)
    {
		this->mat_value = harkio_Matrix_new(harkmatrixtype, src.rows, src.cols);
		if (this->mat_value != NULL){
			this->rows = this->mat_value->rows;
			this->cols = this->mat_value->cols;
			memcpy((void*)this->mat_value->data.complexdata,
			       (void*)src.mat_value->data.complexdata, 
			       sizeof(COMPLEX_TYPE) * this->rows * this->cols);
		}
    }

    Matrix::~Matrix()
    {
		if (mat_value != NULL){
			harkio_Matrix_delete(&mat_value);
			mat_value = NULL;
			this->rows = 0;
			this->cols = 0;
		}
    }
#else
    Matrix::Matrix() : rows(0), cols(0), values(NULL)
    {
    }

    Matrix::Matrix(int rows, int cols) : rows(rows), cols(cols), values(new Complex[rows * cols])
    {
	memset(this->values, 0, sizeof(Complex) * rows * cols);
    }

    Matrix::Matrix(const Matrix &src) : rows(src.rows), cols(src.cols), values(new Complex[src.rows * src.cols])
    {
	memcpy(this->values, src.values, sizeof(Complex) * rows * cols);
    }

    Matrix::~Matrix()
    {
	if (values){
		delete[] values;
		this->values = NULL;
	}
    }
#endif

    Matrix &
    Matrix::operator = (const Matrix &src)
    {
	if(this != &src){
	    Matrix m(src);
	    this->swap(m);
	}
	return *this;
    }

    void
    Matrix::swap(Matrix &src)
    {
	using std::swap;
	swap(this->rows,   src.rows);
	swap(this->cols,   src.cols);
#ifdef USING_HARKIO3_MATRIX
	swap(this->mat_value,   src.mat_value);
#else
	swap(this->values, src.values);
#endif
    }

    int
    Matrix::getRowCount() const
    {
	return this->rows;
    }

    int
    Matrix::getColumnCount() const
    {
	return this->cols;
    }

    void
    Matrix::setElementAt(int row, int col, float rvalue, float ivalue)
    {
	this->setElementAt(row, col, Complex(rvalue, ivalue));
    }

    void
    Matrix::setElementAt(int row, int col, const Complex &value)
    {
	if(this->rows <= row || this->cols <= col){
	    throw IndexOutOfBoundException();
	}

#ifdef USING_HARKIO3_MATRIX
		COMPLEX_TYPE ctemp = value.__rep();
		harkio_Matrix_setValueComplex(this->mat_value, row, col, ctemp);
#else
	this->values[row * this->cols + col] = value;
#endif
    }

#ifdef USING_HARKIO3_MATRIX
    const Complex
#else
    const Complex &
#endif
    Matrix::getElementAt(int row, int col) const
    {
	if(this->rows <= row || this->cols <= col){
	    throw IndexOutOfBoundException();
	}
#ifdef USING_HARKIO3_MATRIX
		COMPLEX_TYPE ctemp;
		harkio_Matrix_getValueComplex(this->mat_value, row, col, &ctemp);
		Complex s = Complex(crealf(ctemp), cimagf(ctemp));
		return s;
#else
	return this->values[row * this->cols + col];
#endif
    }

    const Matrix
    Matrix::getColumn(int col) const
    {
	if(this->cols <= col){
	    throw IndexOutOfBoundException();
	}
#ifdef USING_HARKIO3_MATRIX
		Matrix v;
		v.mat_value = harkio_Matrix_getValues(this->mat_value, 0, this->rows-1, col, col);
		v.rows = v.mat_value->rows;
		v.cols = v.mat_value->cols;
#else
	Matrix v(this->rows, 1);
	for(int i = 0; i < this->rows; i++){
	    v.setElementAt(i, 0, this->getElementAt(i, col));
	}
#endif
	return v;
    }

    void
    Matrix::setColumn(int col, const Matrix &value)
    {
	if(this->cols <= col){
	    throw IndexOutOfBoundException();
	}
	if(value.rows != this->rows){
	    throw MatrixSizeMismatchException();
	}
	if(value.cols != 1){
	    throw NotVectorException();
	}
#ifdef USING_HARKIO3_MATRIX
		harkio_Matrix_setValues(this->mat_value, value.mat_value, 0, col);
#else
	for(int i = 0; i < this->rows; i++){
	    this->setElementAt(i, col, value.getElementAt(i, 0));
	}
#endif
    }

    const Matrix
    Matrix::getRow(int row) const
    {
	if(this->rows <= row){
	    throw IndexOutOfBoundException();
	}
#ifdef USING_HARKIO3_MATRIX
		Matrix v;
		v.mat_value = harkio_Matrix_getValues(this->mat_value, row, row, 0, this->cols-1);
		v.rows = v.mat_value->rows;
		v.cols = v.mat_value->cols;
#else
	Matrix v(1, this->cols);
	for(int i = 0; i < this->cols; i++){
	    v.setElementAt(0, i, this->getElementAt(row, i));
	}
#endif
	return v;
    }

    void
    Matrix::setRow(int row, const Matrix &value)
    {
	if(this->rows <= row){
	    throw IndexOutOfBoundException();
	}
	if(value.cols != this->cols){
	    throw MatrixSizeMismatchException();
	}
	if(value.rows != 1){
	    throw NotVectorException();
	}
#ifdef USING_HARKIO3_MATRIX
		harkio_Matrix_setValues(this->mat_value, value.mat_value, row, 0);
#else
	for(int i = 0; i < this->cols; i++){
	    this->setElementAt(row, i, value.getElementAt(0, i));
	}
#endif
    }

    const Matrix
    Matrix::getSubmatrix(int row, int col, int rows, int cols) const
    {
	if(row < 0 || col < 0 || this->rows < row + rows || this->cols < cols + col){
	    throw IndexOutOfBoundException();
	}
#ifdef USING_HARKIO3_MATRIX
		Matrix v;
		v.mat_value = harkio_Matrix_getValues(this->mat_value, row, row+rows-1, col, col+cols);
		v.rows = v.mat_value->rows;
		v.cols = v.mat_value->cols;
#else
	Matrix v(rows, cols);
	for(int i = 0; i < rows; i++){
	    for(int j = 0; j < cols; j++){
		v.setElementAt(i, j, this->getElementAt(row + i, col + j));
	    }
	}
#endif
	return v;
    }

    void
    Matrix::setSubmatrix(int row, int col, const Matrix &value)
    {
	if(row < 0 || col < 0 || this->rows < row + value.getRowCount() || this->cols < col + value.getColumnCount()){
	    throw IndexOutOfBoundException();
	}
#ifdef USING_HARKIO3_MATRIX
		harkio_Matrix_setValues(this->mat_value, value.mat_value, row, col);
#else
	for(int i = 0; i < value.getRowCount(); i++){
	    for(int j = 0; j < value.getColumnCount(); j++){
		this->setElementAt(row + i, col + j, value.getElementAt(i, j));
	    }
	}
#endif
    }

    const Matrix
    Matrix::diagonalElements() const
    {
	int size = (this->rows < this->cols ? this->rows : this->cols);
#ifdef USING_HARKIO3_MATRIX
		Matrix d;
		d.mat_value = harkio_Matrix_getDiag(this->mat_value);
		d.rows = d.mat_value->rows;
		d.cols = d.mat_value->cols;
#else
	Matrix d(size, 1);

	for(int i = 0; i < size; i++){
	    d.setElementAt(i, 0, this->getElementAt(i, i));
	}
#endif
	return d;
    }

    const Matrix
    Matrix::diagonalMatrix() const
    {
#ifdef USING_HARKIO3_MATRIX
	Matrix d;
	if (this->rows == 1){
		d.mat_value = harkio_Matrix_makeDiag(this->mat_value);
		d.rows = d.mat_value->rows;
		d.cols = d.mat_value->cols;
	}else if(this->cols == 1){
		Matrix m = this->transposed(false);
		d.mat_value = harkio_Matrix_makeDiag(m.mat_value);
		d.rows = d.mat_value->rows;
		d.cols = d.mat_value->cols;
	}
	else {
	    throw NotVectorException();
	}
	return d;
#else
	if(this->rows == 1){
	    Matrix m(this->cols, this->cols);
	    for(int i = 0; i < this->cols; i++){
		for(int j = 0; j < this->cols; j++){
		    m.setElementAt(i, j, (i == j ? this->getElementAt(0, i) : 0.0));
		}
	    }
	    return m;

	}else if(this->cols == 1){
	    Matrix m(this->rows, this->rows);
	    for(int i = 0; i < this->rows; i++){
		for(int j = 0; j < this->rows; j++){
		    m.setElementAt(i, j, (i == j ? this->getElementAt(i, 0) : 0.0));
		}
	    }
	    return m;

	}else{
	    throw NotVectorException();

	}
#endif
    }


    Matrix &
    Matrix::transpose(bool conj)
    {
	Matrix m(this->transposed(conj));
	this->swap(m);
	return *this;
    }

    const Matrix
    Matrix::transposed(bool conj) const
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		if (conj){
			m.mat_value = harkio_Matrix_ctranspose(this->mat_value);
		}
		else{
			m.mat_value = harkio_Matrix_transpose(this->mat_value);
		}
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->cols, this->rows);

	if(conj){
	    for(int i = 0; i < this->rows; i++){
		for(int j = 0; j < this->cols; j++){
		    m.setElementAt(j, i, std::conj(this->getElementAt(i, j)));
		}
	    }
	}else{
	    for(int i = 0; i < this->rows; i++){
		for(int j = 0; j < this->cols; j++){
		    m.setElementAt(j, i, this->getElementAt(i, j));
		}
	    }
	}
#endif
	return m;
    }

	Matrix &
	Matrix::conjugate()
	{
		Matrix m(this->conjugated());
		this->swap(m);
		return *this;
	}
	
	const Matrix
	Matrix::conjugated() const
	{
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_econj(this->mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
		Matrix m(this->rows, this->cols);
		
		for(int i = 0; i < this->rows; i++){
			for(int j = 0; j < this->cols; j++){
				m.setElementAt(i, j, std::conj(this->getElementAt(i, j)));
			}
		}
#endif		
		return m;
	}

    Matrix &
    Matrix::multiply(const Matrix &rhs)
    {
	Matrix m(this->multiplied(rhs));
	this->swap(m);
	return *this;
    }

    const Matrix
    Matrix::multiplied(const Matrix &rhs) const
    {

#ifdef USING_LAPACKE_MATRIX
		if (this->cols != rhs.rows){
	    throw MatrixSizeMismatchException();
		}
		Matrix m(this->rows, rhs.cols);
		Complex alpha(1.0,0.0), beta(0.0,0.0);

//void cblas_cgemm(CBLAS_LAYOUT layout, CBLAS_TRANSPOSE TransA,
//                 CBLAS_TRANSPOSE TransB, const int M, const int N,
//                 const int K, const void *alpha, const void *A,
//                 const int lda, const void *B, const int ldb,
//                 const void *beta, void *C, const int ldc);

		cblas_cgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
								this->rows, rhs.cols, this->cols,
							  (void*)&alpha, this->values, this->cols, 
															rhs.values, rhs.cols,
								(void*)&beta, m.values, m.cols);
#else

	int size = 0;
	if(this->cols != rhs.rows){
	    throw MatrixSizeMismatchException();
	}else{
	    size = this->cols;
	}

#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_mul(this->mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->rows, rhs.cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < rhs.cols; j++){
		Complex v = 0;
		for(int k = 0; k < size; k++){
			v += this->getElementAt(i, k) * rhs.getElementAt(k, j);
		}
		m.setElementAt(i, j, v);
	    }
	}
#endif

#endif

	return m;
    }

    const Matrix
    Matrix::elementwiseSqrt() const
    {
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, std::sqrt(this->getElementAt(i, j)));
	    }
	}
	return m;
    }

    const Matrix
    Matrix::elementwiseAbs() const
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_eabs(this->mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, std::abs(this->getElementAt(i, j)));
	    }
	}
#endif
	return m;
    }

    const Matrix
    Matrix::elementwiseNorm() const
    {
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, std::norm(this->getElementAt(i, j)));
	    }
	}	
	return m;
    }

    const Matrix
    Matrix::elementwiseDivided(const Matrix &rhs) const
    {
	if(this->rows != rhs.rows || this->cols != rhs.cols){
	    throw MatrixSizeMismatchException();
	}
	
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_ediv(this->mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, this->getElementAt(i, j) / rhs.getElementAt(i, j));
	    }
	}
#endif
	return m;
    }

    const Matrix
    Matrix::elementwiseMultiplied(const Matrix &rhs) const
    {
	if(this->rows != rhs.rows || this->cols != rhs.cols){
	    throw MatrixSizeMismatchException();
	}
	
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_emul(this->mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, this->getElementAt(i, j) * rhs.getElementAt(i, j));
	    }
	}
#endif
	return m;
    }

    const Matrix
    Matrix::elementwiseFMultiplied(const Matrix &rhs) const
    {
	if(this->rows != rhs.rows || this->cols != rhs.cols){
	    throw MatrixSizeMismatchException();
	}
	
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		harkio_Matrix *mat = harkio_Matrix_emul(this->mat_value, rhs.mat_value);
		m.mat_value = harkio_Matrix_econj(mat);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
		harkio_Matrix_delete(&mat);
#else
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
	      m.setElementAt(i, j, std::conj(this->getElementAt(i, j)) * rhs.getElementAt(i, j));
	    }
	}
#endif
	return m;
    }

    Complex
    Matrix::getElementSum() const
    {
#ifdef USING_HARKIO3_MATRIX
		COMPLEX_TYPE ctemp;
		harkio_Matrix* mat = harkio_Matrix_sumAll(this->mat_value);
		harkio_Matrix_getValueComplex(mat, 0, 0, &ctemp);
		Complex s(crealf(ctemp), cimagf(ctemp));
		harkio_Matrix_delete(&mat);
#else
	Complex s = 0;
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		s += this->getElementAt(i, j);
	    }
	}
#endif
	return s;
    }

    const Matrix
    Matrix::I(int size)
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_eye(harkmatrixtype, size);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(size, size);
	for(int i = 0; i < size; i++){
	    for(int j = 0; j < size; j++){
		m.setElementAt(i, j, (i == j ? 1.0 : 0.0));
	    }
	}
#endif
	return m;
    }

    const Matrix
    Matrix::O(int rows, int cols)
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		if (rows == 0 || cols == 0){
			return m;
		}
		m.mat_value = harkio_Matrix_zeros(harkmatrixtype, rows, cols);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(rows, cols);
	for(int i = 0; i < rows; i++){
	    for(int j = 0; j < cols; j++){
		m.setElementAt(i, j, 0);
	    }
	}
#endif
	return m;
    }

    Complex
    Matrix::scalar() const
    {
	if(this->rows != 1 || this->cols != 1){
	    throw NotScalarException();
	}
	return this->getElementAt(0, 0);
    }

    Complex
    Matrix::norm() const
    {

#ifdef USING_LAPACKE_MATRIX
	float v;
	if ((this->cols>1) && (this->rows>1)){
	    throw NotVectorException();
	}

//float LAPACKE_zlange( int matrix_layout, char norm, lapack_int m,
//                       lapack_int n, const lapack_complex_float* a,
//                       lapack_int lda );

	v = LAPACKE_clange(LAPACK_ROW_MAJOR, 'F', this->rows, this->cols, (const lapack_complex_float*)this->values, this->cols);	
	return Complex(v, 0.0);
#else

	Complex n = 0;
	if(this->rows == 1){
	    for(int i = 0; i < this->cols; i++){
		Complex v = this->getElementAt(0, i);
		n += v * conj(v);
	    }
	}else if(this->cols == 1){
	    for(int i = 0; i < this->rows; i++){
		Complex v = this->getElementAt(i, 0);
		n += v * conj(v);
	    }
	}else{
	    throw NotVectorException();
	}
	return std::sqrt(n);
#endif

    }

	Complex
	Matrix::frobeniusNorm() const
	{

#ifdef USING_LAPACKE_MATRIX
	float v;

//float LAPACKE_zlange( int matrix_layout, char norm, lapack_int m,
//                       lapack_int n, const lapack_complex_float* a,
//                       lapack_int lda );

	v = LAPACKE_clange(LAPACK_ROW_MAJOR, 'F', this->rows, this->cols, (const lapack_complex_float*)this->values, this->cols);	
	return Complex(v, 0.0);
#else

		Complex n = 0;
		for(int i = 0; i < this->rows; i++){
			for(int j = 0; j < this->cols; j++){
				Complex v = this->getElementAt(i, j);
				n += v * conj(v);
			}
		}
		return std::sqrt(n);

#endif

	}

    const Matrix
    Matrix::absoluteValues() const
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_eabs(this->mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		m.setElementAt(i, j, std::abs(this->getElementAt(i, j)));
	    }
	}
#endif
	return m;
    }

    const Matrix
    Matrix::pseudoInverse() const
    {
	Matrix U, S, V ;
	float a,b;
	if(this->cols > this->rows){
		
		Matrix A = this->transposed(false);
		A.singularValueDecomposition(U, S, V);

		Matrix SS(A.cols, A.rows);
		for(int i = 0; i < A.rows && i < A.cols; i++){
		    Complex v = S.getElementAt(i, 0);
		    if(abs(v) > HARKCORE::H_EPS){
				//std::complex<float> temp= 1.0 / v;
				a=v.real(); b=v.imag();
				std::complex<float> temp(a/(a*a+b*b), -b/(a*a+b*b));
			SS.setElementAt(i, i, temp);
		    }
		}
		return (V * SS * U.transposed()).transposed(false);
	}
	else{
		this->singularValueDecomposition(U, S, V);

		Matrix SS(this->cols, this->rows);
		for(int i = 0; i < this->rows && i < this->cols; i++){
		    Complex v = S.getElementAt(i, 0);
		    if(abs(v) > HARKCORE::H_EPS){
				a=v.real(); b=v.imag();
				std::complex<float> temp(a/(a*a+b*b), -b/(a*a+b*b));
				SS.setElementAt(i, i, temp);
		    }
		}
		return V * SS * U.transposed();
	}	
    }

    const Matrix
    Matrix::Inverse() const
    {
#ifdef USING_LAPACKE_MATRIX
		int *ipiv;
		int rtn;
	
		if (this->cols != this->rows){
			throw NotSquareMatrixException();
		}

		ipiv = new int[this->rows];

		// BX=Iを解く
		Matrix m = I(this->rows);
		Matrix a(*this);	// aは破壊されるので元データはコピーして使用
/*
		lapack_int LAPACKE_zgesv_work( int matrix_layout, lapack_int n, lapack_int nrhs,
                               lapack_complex_float* a, lapack_int lda,
                               lapack_int* ipiv, lapack_complex_float* b,
                               lapack_int ldb );*/
		rtn = LAPACKE_cgesv_work(LAPACK_ROW_MAJOR, a.rows, a.cols, (lapack_complex_float*)a.values, a.cols, ipiv, (lapack_complex_float*)m.values, m.cols);
		if (rtn != 0){
			throw FailedLAPACKEException();
		}

		delete [] ipiv;	
		return m;
#else
        boost::numeric::ublas::matrix<std::complex<float> > a(this->rows, this->cols), c(this->rows, this->cols);

	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		a(i, j) = this->getElementAt(i, j);
	    }
	}
  
	int i,j,k=0;
	int p;

	boost::numeric::ublas::matrix<std::complex<float> > tmpa(a.size1(), a.size2());
  
	std::complex<float> x,y=0.0;
	std::complex<float> tmpfc = 0.0;
	std::complex<float> tmpfc2 = 0.0;
	float tmp1 = 0.0;
	float tmp2 = 0.0;

	noalias(tmpa) = a;

	p = tmpa.size1();

	if(p == tmpa.size2()){
	  
	  // Make a Unit Matrix
	  for(i=0;i<p;i++){
	    for(j=0;j<p;j++){
	      if(i==j){
		c(i,j) = std::complex<float>(1.0,0.0);
	      }else{
		c(i,j) = std::complex<float>(0.0,0.0);
	      }
	    }
	  }
	  for(i=0;i<p;i++){
	    if((tmpa(i,i).real()==0.0f)&&(tmpa(i,i).imag()==0.0f)){
	      for(j=i;j<p;j++){
		if((tmpa(j,i).real()!=0.0f)||(tmpa(j,i).imag()!=0.0f)){
		  for(k=0;k<p;k++){
		    y=tmpa(i,k);
		    tmpa(i,k)=tmpa(j,k);
		    tmpa(j,k)=y;
		    y=c(i,k);
		    c(i,k)=c(j,k);
		    c(j,k)=y;
		  }
		  k=1;
		  break;
		};
	      }
	      if(k!=1)
		printf("ERROR!!No inverse matrix available.\n");
	    }
	    for(j=0;j<p;j++){
	      tmp1 = (tmpa(j,i).real()*tmpa(i,i).real()+tmpa(j,i).imag()*tmpa(i,i).imag())/(tmpa(i,i).real()*tmpa(i,i).real()+tmpa(i,i).imag()*tmpa(i,i).imag());
	      tmp2 = (tmpa(j,i).imag()*tmpa(i,i).real()-tmpa(j,i).real()*tmpa(i,i).imag())/(tmpa(i,i).real()*tmpa(i,i).real()+tmpa(i,i).imag()*tmpa(i,i).imag());
	      tmpfc = std::complex<float>(tmp1,tmp2);
	      x = -tmpfc;
	      if(i==j)
		{
		  x = std::complex<float>(0.0,0.0);
		}
	      for(k=0;k<p;k++){
		tmp1 = x.real()*c(i,k).real() - x.imag()*c(i,k).imag();
		tmp2 = x.real()*c(i,k).imag() + x.imag()*c(i,k).real();
	        tmpfc = std::complex<float>(tmp1,tmp2);
		tmp1 = c(j,k).real() + tmpfc.real();
		tmp2 = c(j,k).imag() + tmpfc.imag();
		tmpfc2 = std::complex<float>(tmp1,tmp2);
		c(j,k) = tmpfc2;
		tmp1 = x.real()*tmpa(i,k).real() - x.imag()*tmpa(i,k).imag();
		tmp2 = x.real()*tmpa(i,k).imag() + x.imag()*tmpa(i,k).real();
		tmpfc = std::complex<float>(tmp1,tmp2);
		tmp1 = tmpa(j,k).real() + tmpfc.real();
		tmp2 = tmpa(j,k).imag() + tmpfc.imag();
		tmpfc2 = std::complex<float>(tmp1,tmp2);
		tmpa(j,k) = tmpfc2;
	      }
	    }
	    x=tmpa(i,i);
	    for(k=0;k<p;k++){
	      tmp1 = (tmpa(i,k).real()*x.real()+tmpa(i,k).imag()*x.imag())/(x.real()*x.real()+x.imag()*x.imag());
	      tmp2 = (tmpa(i,k).imag()*x.real()-tmpa(i,k).real()*x.imag())/(x.real()*x.real()+x.imag()*x.imag());
	      tmpfc = std::complex<float>(tmp1,tmp2);
	      tmpa(i,k) = tmpfc;
	      tmp1 = (c(i,k).real()*x.real()+c(i,k).imag()*x.imag())/(x.real()*x.real()+x.imag()*x.imag());
	      tmp2 = (c(i,k).imag()*x.real()-c(i,k).real()*x.imag())/(x.real()*x.real()+x.imag()*x.imag());
	      tmpfc = std::complex<float>(tmp1,tmp2);
	      c(i,k) = tmpfc;
	    }
	  } 
	  
	}else{
	  printf("Error : Inverse : Not a square matrix.\n");
	}

	Matrix OUT(this->rows, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		OUT.setElementAt(i, j, c(i, j));
	    }
	}
	return OUT;
#endif	
    }


    void
    Matrix::singularValueDecomposition(Matrix &u, Matrix &s, Matrix &v) const
    {
#ifdef USING_LAPACKE_MATRIX
			float *vec_s;
			int rtn;

			int ssize = this->rows > this->cols ? this->rows : this->cols;
			int ssize_lpk = this->rows < this->cols ? this->rows : this->cols;
			Matrix a(*this);
			Matrix U(this->rows, this->rows), S(ssize, 1), V(this->cols, this->cols);
	
			vec_s = new float[ssize_lpk];

			// SVD
			/*
			lapack_int LAPACKE_zgesdd( int matrix_layout, char jobz, 
													 lapack_int m, lapack_int n, lapack_complex_float* a,
                           lapack_int lda, float* s, lapack_complex_float* u,
                           lapack_int ldu, lapack_complex_float* vt,
                           lapack_int ldvt );
			*/
			rtn = LAPACKE_cgesdd(LAPACK_ROW_MAJOR, 'A', a.rows, a.cols, (lapack_complex_float*)a.values, a.cols, vec_s, (lapack_complex_float*)U.values, U.cols, (lapack_complex_float*)V.values, V.cols);

			if (rtn != 0){
				delete [] vec_s;
				throw FailedLAPACKEException();
			}

			for (int i = 0 ; i < ssize_lpk ; i++){
				S.setElementAt(i,0, Complex(vec_s[i],0.0));
			}
			u.swap(U);
			v.swap(V.transpose(true));
			s.swap(S);

			delete [] vec_s;
			
#else
	int ssize = this->rows > this->cols ? this->rows : this->cols;
	boost::numeric::ublas::matrix<std::complex<float> > bA(this->rows, this->cols), bU(this->rows, this->rows), bV(this->cols, this->cols);
	boost::numeric::ublas::vector<float> bS(ssize);

	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		bA(i, j) = this->getElementAt(i, j);
	    }
	}
	svd(bA, bS, bU, bV);

	Matrix U(this->rows, this->rows), S(ssize, 1), V(this->cols, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->rows; j++){
		U.setElementAt(i, j, bU(i, j));
	    }
	}
	for(int i = 0; i < bS.size(); i++){
	    S.setElementAt(i, 0, bS(i));
	}
	for(int i = 0; i < this->cols; i++){
	    for(int j = 0; j < this->cols; j++){
		V.setElementAt(i, j, bV(i, j));
	    }
	}

	u.swap(U);
	s.swap(S);
	v.swap(V);
#endif
    }

    void
    Matrix::eigenValues(Matrix &eigenValues, Matrix &eigenVectors) const
    {
#ifdef USING_LAPACKE_MATRIX
			int rtn;
			float *vec_d;

			Matrix a(*this), EV(this->rows, 1), ED(this->rows, this->rows);

			if (this->rows != this->cols){
				throw NotSquareMatrixException();		
			}

			vec_d = new float[a.rows];
			// EVD
			/*
			lapack_int LAPACKE_zheevd( int matrix_layout, char jobz, char uplo, lapack_int n,
                           lapack_complex_float* a, lapack_int lda, float* w );	
			*/
			rtn = LAPACKE_cheevd(LAPACK_ROW_MAJOR, 'V', 'U', a.rows, (lapack_complex_float*)a.values, a.cols, vec_d);

			if (rtn != 0){
				delete [] vec_d;
				throw FailedLAPACKEException();
			}

			for (int i = 0 ; i < this->rows ; i++){
				EV.setElementAt(i,0, Complex(vec_d[i],0.0));
			}

			delete [] vec_d;

			eigenValues.swap(EV);
			eigenVectors.swap(a);
#else
	if(this->rows != this->cols){
	    throw NotSquareMatrixException();
	}
	int size = this->rows;

	boost::numeric::ublas::matrix<std::complex<float> > bA(size, size), bV(size, size);
	boost::numeric::ublas::vector<float> bD(size);

//	boost::numeric::ublas::matrix<std::complex<float> > bAsave(size, size);
	for(int i = 0; i < size; i++){
	    for(int j = 0; j < size; j++){
		bA(i, j) = this->getElementAt(i, j);
		//bAsave(i, j) = bA(i, j) = this->getElementAt(i, j);
	    }
	}
	//std::cout << bA << std::endl;
	evd(bA, bD, bV);
	//std::cout << bD << std::endl;
	//std::cout << bV << std::endl;
	//std::cout << prod(bAsave, bV) << std::endl;

//	boost::numeric::ublas::matrix<std::complex<float> > bDD(size, size);
//	for(int i = 0; i < size; i++){
//	    for(int j = 0; j < size; j++){
//		bDD(i, j) = ((i == j) ? bD(i) : 0.0);
//	    }
//	}
	//std::cout << bDD << std::endl;
	//std::cout << prod(bV, bDD) << std::endl;

	Matrix V(size, size), D(size, 1);
	for(int i = 0; i < size; i++){
	    for(int j = 0; j < size; j++){
		V.setElementAt(i, j, bV(i, j));
	    }
	}
	for(int i = 0; i < size; i++){
	    D.setElementAt(i, 0, bD(i));
	}

	V.swap(eigenVectors);
	D.swap(eigenValues);
#endif
    }

    /* 正定値エルミート行列のコレスキー分解を行う。 */
    const Matrix
    Matrix::choleskyDecompositionForPositiveDefiniteHermitianMatrix()
    {
#ifdef USING_LAPACKE_MATRIX
			Matrix M(*this);
			int rtn;
			rtn = LAPACKE_cpotrf(LAPACK_ROW_MAJOR, 'U', M.rows, (lapack_complex_float*)M.values, M.cols);
			if (rtn != 0){
				throw FailedLAPACKEException();
			}
			for (int i = 0 ; i < M.rows ; i++){	// 下半分に0
				for (int j = 0 ; j < i ; j++){
					M.setElementAt(i, j, Complex(0.0,0.0));
				}
			}
			return M;
#else
	if(this->rows != this->cols){
	    throw NotSquareMatrixException();
	}
	int size = this->rows;
	
	Matrix M(size, size);
	for(int i = 0; i < size; i++){
	    Complex v = this->getElementAt(i, i);

	    for(int j = 0; j < i; j++){
		v -= M.getElementAt(j, i) * conj(M.getElementAt(j, i));
	    }
	    M.setElementAt(i, i, sqrt(v));

	    for(int j = i + 1; j < size; j++){
		Complex v = this->getElementAt(i, j);
		for(int k = 0; k < i; k++){
		    v -= M.getElementAt(k, j) * conj(M.getElementAt(k, i));
		}
		M.setElementAt(i, j, v / M.getElementAt(i, i));
	    }
	}

	return M;
#endif
    }


    
    /* 正定値エルミート行列のコレスキー分解を行う。 */
/*
    static Matrix
    choleskyDecompositionForPositiveDefiniteHermitianMatrix(const Matrix &m)
    {
	if(m.getRowCount() != m.getColumnCount()){
	    throw NotSquareMatrixException();
	}
	int size = m.getRowCount();
	
	Matrix M(size, size);
	for(int i = 0; i < size; i++){
	    Complex v = m.getElementAt(i, i);

	    for(int j = 0; j < i; j++){
		v -= M.getElementAt(j, i) * conj(M.getElementAt(j, i));
	    }
	    M.setElementAt(i, i, sqrt(v));

	    for(int j = i + 1; j < size; j++){
		Complex v = m.getElementAt(i, j);
		for(int k = 0; k < i; k++){
		    v -= M.getElementAt(k, j) * conj(M.getElementAt(k, i));
		}
		M.setElementAt(i, j, v / M.getElementAt(i, i));
	    }
	}

	return M;
    }
*/
    void
    Matrix::generalizedEigenValues(const Matrix &mat, Matrix &eigenValues, Matrix &eigenVectors) const
    {
	Matrix m(mat);
	Matrix r = m.choleskyDecompositionForPositiveDefiniteHermitianMatrix();
	
//	this->dump(stdout, "A");
//	mat.dump(stdout, "m");
//r.dump(stdout, "r");
//	(r * r.transposed()).dump(stdout, "rr'");
//	(r.transposed() * r).dump(stdout, "r'r");
	
	Matrix ri = r.pseudoInverse();
//ri.dump(stdout, "rinv");
//	(ri *r).dump(stdout, "rinvr");
//	(r *ri).dump(stdout, "rrinv");
	Matrix AA = ri.transposed() * (*this) * ri;
//(ri.transposed()).dump(stdout, "1");
//(ri.transposed() * (*this)).dump(stdout, "2");
//(ri.transposed() * (*this) * ri).dump(stdout, "3");
//AA.dump(stdout, "(rinv)'mr");
	Matrix VV, D;
	AA.eigenValues(D, VV);
//D.dump(stdout, "D");
//VV.dump(stdout, "VV");

//Matrix AA_VV = AA*VV;
//Matrix VV_D = VV*D.diagonalMatrix();
//AA_VV.dump(stdout, "AA_VV");
//VV_D.dump(stdout, "VV_DD");

	Matrix V = ri * VV;
	//V.dump(stdout, "V");
	
	V.swap(eigenVectors);
	D.swap(eigenValues);
    }

    void
    Matrix::generalizedSingularValues(const Matrix &K, Matrix &u, Matrix &s, Matrix &v) const
    {
#ifdef USING_LAPACKE_MATRIX
			Matrix Kinv = K.Inverse();
			Matrix KinvR = Kinv * (*this);
			KinvR.singularValueDecomposition(u, s, v);
#else
	int ssize = this->rows > this->cols ? this->rows : this->cols;
	boost::numeric::ublas::matrix<std::complex<float> > bA(this->rows, this->cols), bU(this->rows, this->rows), bV(this->cols, this->cols);
	boost::numeric::ublas::vector<float> bS(ssize);

	Matrix Kinv = K.Inverse();
	Matrix KinvR = Kinv * (*this);

	for(int i = 0; i < KinvR.rows; i++){
	    for(int j = 0; j < KinvR.cols; j++){
		bA(i, j) = KinvR.getElementAt(i, j);
	    }
	}
	svd(bA, bS, bU, bV);

	Matrix U(this->rows, this->rows), S(ssize, 1), V(this->cols, this->cols);
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->rows; j++){
		U.setElementAt(i, j, bU(i, j));
	    }
	}
	for(int i = 0; i < bS.size(); i++){
	    S.setElementAt(i, 0, bS(i));
	}
	for(int i = 0; i < this->cols; i++){
	    for(int j = 0; j < this->cols; j++){
		V.setElementAt(i, j, bV(i, j));
	    }
	}

	u.swap(U);
	s.swap(S);
	v.swap(V);
#endif
    }

    Complex
    Matrix::getMaximumValue() const
    {
	Complex m = 0;
	float  n = 0;
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		Complex v = this->getElementAt(i, j);
		if(n < std::abs(v)){
		    m = v;
		    n = std::abs(v);
		}
	    }
	}
	return m;
    }

    Complex
    Matrix::getMinimumValue() const
    {
	Complex m = DBL_MAX;
	float  n = DBL_MAX;
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		Complex v = this->getElementAt(i, j);
		if(n > std::abs(v)){
		    m = v;
		    n = std::abs(v);
		}
	    }
	}
	return m;
    }

    int
    Matrix::getMaximumValueIndexInRow(int row) const
    {
	int maxi;
	float maxn;
	for(int i = 0; i < this->cols; i++){
	    Complex v = this->getElementAt(row, i);
	    if(i == 0 || maxn < std::abs(v)){
		maxi = i;
		maxn = std::abs(v);
	    }
	}
	return maxi;
    }

    int
    Matrix::getMaximumValueIndexInColumn(int col) const
    {
	int maxi;
	float maxn;
	for(int i = 0; i < this->rows; i++){
	    Complex v = this->getElementAt(i, col);
	    if(i == 0 || maxn < std::abs(v)){
		maxi = i;
		maxn = std::abs(v);
	    }
	}
	return maxi;
    }

    void
    Matrix::dump(FILE *stream) const
    {
	this->dump(stream, "Matrix: %p");
    }

    void
    Matrix::dump(FILE *stream, const char *format, ...) const
    {
	va_list args;
	va_start(args, format);
	vfprintf(stream, format, args);
	va_end(args);


	fprintf(stream, ": rows=%u, cols=%u\n", this->rows, this->cols);
	fprintf(stream, "[ ");
	for(int i = 0; i < this->rows; i++){
	    for(int j = 0; j < this->cols; j++){
		Complex v = this->getElementAt(i, j);
		fprintf(stream, "%.6f%+.6fi ", v.real(), v.imag());
	    }
	    fprintf(stream, "\n");
	}
	fprintf(stream, "]\n");
    }

    const Matrix
    operator + (const Matrix &lhs, const Matrix &rhs)
    {
	if(lhs.getRowCount() != rhs.getRowCount() || lhs.getColumnCount() != rhs.getColumnCount()){
	    throw;
	}

#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_add(lhs.mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(lhs);
	for(int i = 0; i < lhs.getRowCount(); i++){
	    for(int j = 0; j < lhs.getColumnCount(); j++){
		m.setElementAt(i, j, m.getElementAt(i, j) + rhs.getElementAt(i, j));
	    }
	}
#endif
	return m;
    }

    const Matrix
    operator - (const Matrix &lhs, const Matrix &rhs)
    {
	if(lhs.getRowCount() != rhs.getRowCount() || lhs.getColumnCount() != rhs.getColumnCount()){
	    throw;
	}

#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_sub(lhs.mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(lhs);
	for(int i = 0; i < lhs.getRowCount(); i++){
	    for(int j = 0; j < lhs.getColumnCount(); j++){
		m.setElementAt(i, j, m.getElementAt(i, j) - rhs.getElementAt(i, j));
	    }
	}
#endif
	return m;
    }

    const Matrix
    operator * (const Matrix &lhs, const Matrix &rhs)
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		m.mat_value = harkio_Matrix_mul(lhs.mat_value, rhs.mat_value);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
	return m;
#else
	return lhs.multiplied(rhs);
#endif
    }

    const Matrix
    operator * (const Complex &lhs, const Matrix &rhs)
    {
	return rhs * lhs;
    }

    const Matrix
    operator * (const Matrix &lhs, const Complex &rhs)
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		COMPLEX_TYPE crhs = rhs.__rep();
		m.mat_value = harkio_Matrix_emulCComplex(lhs.mat_value, crhs);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(lhs);
	for(int i = 0; i < m.getRowCount(); i++){
	    for(int j = 0; j < m.getColumnCount(); j++){
		m.setElementAt(i, j, m.getElementAt(i, j) * rhs);
	    }
	}
#endif
	return m;
    }
    
    const Matrix
    operator / (const Matrix &lhs, const Complex &rhs)
    {
#ifdef USING_HARKIO3_MATRIX
		Matrix m;
		COMPLEX_TYPE crhs = rhs.__rep();
		m.mat_value = harkio_Matrix_edivCComplex(lhs.mat_value, crhs);
		m.rows = m.mat_value->rows;
		m.cols = m.mat_value->cols;
#else
	Matrix m(lhs);
	for(int i = 0; i < m.getRowCount(); i++){
	    for(int j = 0; j < m.getColumnCount(); j++){
		m.setElementAt(i, j, m.getElementAt(i, j) / rhs);
	    }
	}
#endif
	return m;
    }
}

namespace std
{
    template<> void swap<HARKCORE::Matrix>(HARKCORE::Matrix &a, HARKCORE::Matrix &b)
    {
	a.swap(b);
    }
}

int main(){

	HARKCORE::Matrix G(5, 5);
	for(int i=0; i<4; i++){
		for(int j=0; j<4; j++){
			G.setElementAt(i, j, i*4+j+1);
		}
	}
	G.setElementAt(0, 4, -11);
	G.setElementAt(1, 4, -5);
	G.setElementAt(2, 4, -6);
	G.setElementAt(3, 4, -7);
	G.setElementAt(3, 1, -14);
	G.setElementAt(4, 0, -2);
	G.setElementAt(4, 1, -8);
	G.setElementAt(4, 2, 5);
	G.setElementAt(4, 3, 10);
	G.setElementAt(4, 4, 45);
	HARKCORE::Matrix A(5, 5), B(5, 5), C(5, 5);
	G.singularValueDecomposition(A, B, C);
	G.dump(stdout,"G");
	A.dump(stdout,"A");
	B.dump(stdout,"B");
	C.dump(stdout,"C");

	HARKCORE::Matrix O(5, 5);
	for(int i=0; i<4; i++){
		O.setElementAt(i, i,B.getElementAt(i,0));
	}
	O.dump(stdout,"O");
	HARKCORE::Matrix tt(5, 5);
	std::cout << "new" << std::endl;
	HARKCORE::Matrix tt2(5, 5);
	std::cout << "new2" << std::endl;
	//tt = A.transposed() ;
	tt.dump(stdout,"tt");
	std::cout << "tt" << std::endl;
	tt2 = A * O * C.transposed();
	std::cout << "tt2" << std::endl;
	tt2.dump(stdout,"A*B*C'");


}
