#include <string>
#include <iostream>
#include <cstdlib>
#include <complex>
#include <cmath>

#include "BufferedNode.h"
#include "Vector.h"
#include "Matrix.h"
#include "Map.h"
#include "Source.h"

#include "BF.h"
#include "BF_Matrix.h"
#include "BF_DS.h"
#include "BF_WDS.h"
#include "BF_NULL.h"
#include "BF_ILSE.h"
#include "BF_LCMV.h"
#include "BF_GJ.h"
#include "BF_GSS.h"
#include "BF_GICA.h"

#if 0 
#ifdef GSS_ENABLE
 #include "BF_GHDSS.h"
#endif
#endif

#include "BF_GHDSS.h"

#include "ArrayInit.h"
#include "MICARY_struct.h"
//#include <../config.h>
#include <iostream>

using namespace std;
using namespace FD;

#ifdef ENABLE_HARKIO3
	#include "HarkFileIO_NH.h"
#elif  ENABLE_HARKIO2
	#include "HarkFileIO.h"
	using namespace harkio2;
#elif ENABLE_HARKIO1
	#include "HarkFileIO.h"
	using namespace harkio;
#endif

#include "HarkFileIOByMatrix.h"


class BeamForming;

DECLARE_NODE(BeamForming);
/*Node
 *
 * @name BeamForming
 * @category HARK:Separation
 * @description Execute sound source separation with BeamForming.
 *
 * @input_name INPUT_FRAMES
 * @input_type Matrix<complex<float> >
 * @input_description Input multi-channel spectrum. A row is a channel, and a column is a spectrum for the corresponding channel. 
 *
 * @input_name INPUT_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @input_name INPUT_NOISE_SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Noise source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description Separated spectrum with a source ID(key), and the value is the separated spectrum for each sound source (Vector<complex<float> >).
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description The frame length of each channel (in samples) [default: 512].
 *
 * @parameter_name ADVANCE
 * @parameter_type int
 * @parameter_value 160
 * @parameter_description The shift length beween adjacent frames (in samples)[default: 160].
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description Sampling rate (Hz) [default: 16000].
 *
 * @parameter_name LOWER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description Lower bound of frequency (Hz). [default: 0]
 * 
 * @parameter_name UPPER_BOUND_FREQUENCY
 * @parameter_type int
 * @parameter_value 8000
 * @parameter_description Upper bound of frequency (Hz). [default: 8000]
 * 
 * @parameter_name TF_CONJ_FILENAME
 * @parameter_type string
 * @parameter_description Filename of a pre-measured transfer function for separation.
 * 
 * @parameter_name INITW_FILENAME
 * @parameter_type string
 * @parameter_valid BF_METHOD=GICA:LCMV:GJ:GHDSS
 * @parameter_description Filename of an initial separation matrix. If specified, a matrix in INITW_FILENAME is used as an initial separation matrix. Otherwise, initial separation matrix is estimated from the geometrical relationship or pre-measured TF according to TF_CONJ.
 *
 * @parameter_name GICA_SS_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=GICA
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:LC_MYU:ADAPTIVE
 * @parameter_description The calculation method for SS step size parameter corresponding to the blind separation part. "FIX" uses a fixed step size as GICA_SS_MYU, "LC_MYU" uses the same value as GICA_LC_MYU, and "ADAPTIVE" adaptively estimates an optimal step size. This is used only when BF_METHOD=GICA. [default: ADAPTIVE]
 *
 * @parameter_name GHDSS_SS_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:LC_MYU:ADAPTIVE
 * @parameter_description The calculation method for SS step size parameter corresponding to the blind separation part. "FIX" uses a fixed step size as GHDSS_SS_MYU, "LC_MYU" uses the same value as GHDSS_LC_MYU, and "ADAPTIVE" adaptively estimates an optimal step size. This is used only when BF_METHOD=GHDSS. [default: ADAPTIVE]
 *
 * @parameter_name GICA_SS_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=GICA
 * @parameter_value 0.001
 * @parameter_description SS step size value if GICA_SS_METHOD=FIX. SS adaptive step size scaler if GICA_SS_METHOD=ADAPTIVE. [default 0.001]
 *
 * @parameter_name GHDSS_SS_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_value 0.001
 * @parameter_description SS step size value if GHDSS_SS_METHOD=FIX. SS adaptive step size scaler if GHDSS_SS_METHOD=ADAPTIVE. [default 0.001]
 *
 * @parameter_name LCMV_LC_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=LCMV
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:ADAPTIVE
 * @parameter_description The calculation method for LC step size corresponding to geometric constraints. "FIX" uses a fixed value as LCMV_LC_MYU, and "Adaptive" adaptively estimates an optimal step size as LCMV_LC_MYU. This is used only when BF_METHOD=LCMV. [default: ADAPTIVE]
 *
 * @parameter_name GJ_LC_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=GJ
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:ADAPTIVE
 * @parameter_description The calculation method for LC step size corresponding to geometric constraints. "FIX" uses a fixed value as GJ_LC_MYU, and "Adaptive" adaptively estimates an optimal step size as GJ_LC_MYU. This is used only when BF_METHOD=GJ. [default: ADAPTIVE]
 *
 * @parameter_name GICA_LC_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=GICA
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:ADAPTIVE
 * @parameter_description The calculation method for LC step size corresponding to geometric constraints. "FIX" uses a fixed value as GICA_LC_MYU, and "Adaptive" adaptively estimates an optimal step size as GICA_LC_MYU. This is used only when BF_METHOD=GICA. [default: ADAPTIVE]
 *
 * @parameter_name GHDSS_LC_METHOD
 * @parameter_type string
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_value ADAPTIVE
 * @parameter_list FIX:ADAPTIVE
 * @parameter_description The calculation method for LC step size corresponding to geometric constraints. "FIX" uses a fixed value as GHDSS_LC_MYU, and "Adaptive" adaptively estimates an optimal step size as GHDSS_LC_MYU. This is used only when BF_METHOD=GHDSS. [default: ADAPTIVE]
 *
 * @parameter_name LCMV_LC_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=LCMV
 * @parameter_value 0.001
 * @parameter_description LC step size value if LCMV_LC_METHOD=FIX. LC adaptive step size scaler if LCMV_LC_METHOD=ADAPTIVE. [default 0.001]
 *
 * @parameter_name GJ_LC_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=GJ
 * @parameter_value 0.001
 * @parameter_description LC step size value if GJ_LC_METHOD=FIX. LC adaptive step size scaler if GJ_LC_METHOD=ADAPTIVE. [default 0.001]
 *
 * @parameter_name GICA_LC_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=GICA
 * @parameter_value 0.001
 * @parameter_description LC step size value if GICA_LC_METHOD=FIX. LC adaptive step size scaler if GICA_LC_METHOD=ADAPTIVE and if GICA_SS_METHOD=LC_MYU. [default 0.001]
 *
 * @parameter_name GHDSS_LC_MYU
 * @parameter_type float
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_value 0.001
 * @parameter_description LC step size value if GHDSS_LC_METHOD=FIX. LC adaptive step size scaler if GHDSS_LC_METHOD=ADAPTIVE and if GHDSS_SS_METHOD=LC_MYU. [default 0.001]
 *
 * @parameter_name GHDSS_LC_CONST
 * @parameter_type string
 * @parameter_value FULL
 * @parameter_list FULL:DIAG
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_description The calculation method for geometric constraints. "FULL" uses all elements of a matrix, and "DIAG" only uses diagonal parts. [default: FULL]
 *
 * @parameter_name GICA_SS_SCAL
 * @parameter_type float
 * @parameter_valid BF_METHOD=GICA
 * @parameter_value 1.0
 * @parameter_description Scaling factor of the nonlinear function for SS step size. [default: 1.0]
 *
 * @parameter_name GHDSS_SS_SCAL
 * @parameter_type float
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_value 1.0
 * @parameter_description Scaling factor of the nonlinear function for SS step size. [default: 1.0]
 *
 * @parameter_name GHDSS_NOISE_FLOOR
 * @parameter_type float
 * @parameter_value 0.0
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_description Noise floor value. [default 0.0]
 *
 * @parameter_name GHDSS_UPDATE
 * @parameter_type string
 * @parameter_value STEP
 * @parameter_list STEP:TOTAL
 * @parameter_valid BF_METHOD=GHDSS
 * @parameter_description The update method of separation matrix. "STEP" updates W sequentially, i.e., based on SS and then on LC cost. "TOTAL" updates W based on an integrated value of SS and LC cost [default: STEP]
 *
 * @parameter_name UPDATE_METHOD_W
 * @parameter_type string
 * @parameter_value ID
 * @parameter_list ID:POS:ID_POS
 * @parameter_valid BF_METHOD=GICA:LCMV:GJ:GHDSS
 * @parameter_description Switching method of separation matrix, W. [default: ID]
 *
 * @parameter_name UPDATE_ACCEPT_DISTANCE
 * @parameter_type float
 * @parameter_value 300.0
 * @parameter_valid BF_METHOD=GICA:LCMV:GJ:GHDSS
 * @parameter_description Angle allowance to switch separation matrix in mm. This parameter is available when UPDATE_METHOD_W is POS or ID_POS and when INITW_FILENAME is set. [default: 300.0]
 *
 * @parameter_name EXPORT_W
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_valid BF_METHOD=GICA:LCMV:GJ:GHDSS
 * @parameter_description Separation matrix W is exported if true. [default: false]
 *
 * @parameter_name EXPORT_W_FILENAME
 * @parameter_type string
 * @parameter_valid BF_METHOD=GICA:LCMV:GJ:GHDSS
 * @parameter_description The filename to export W.
 *
 * @parameter_name BF_METHOD
 * @parameter_type string
 * @parameter_value LCMV
 * @parameter_list DS:WDS:NULL:ILSE:LCMV:GJ:GICA:GHDSS
 * @parameter_description beam-forming algorithm type.
 *
 * @parameter_name ENABLE_DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description enable debug print
 *
 END*/

class BeamForming : public BufferedNode
{
public:
    BeamForming(string nodeName, ParameterSet params)
	: BufferedNode(nodeName, params),
	  ssScalingFactor(0.1),
	  ssMyu(0.001),
	  lcMyu(0.001)
    {
	/*== FlowDesigner 入出力インタフェース ==*/

	/* 入出力IDの割り当て */
	this->input_frameID        = addInput("INPUT_FRAMES");
	this->input_signalSourceID = addInput("INPUT_SOURCES");
	this->input_noiseSourceID  = -1;//addInput("INPUT_NOISE_SOURCES");
	this->output_ID            = addOutput("OUTPUT");

	/*== パラメータの取得と検証 ==*/

	/* ビームフォーミング手法の選択 */
	string bfMethod = object_cast<string>(parameters.get("BF_METHOD"));
	if (bfMethod == "DS") {
	    this->bfType = BF::BfDS;
	} else if(bfMethod == "WDS") {
	    this->bfType = BF::BfWDS;
	} else if(bfMethod == "NULL") {
	    this->bfType = BF::BfNull;
	} else if(bfMethod == "ILSE") {
	    this->bfType = BF::BfILSE;
	} else if(bfMethod == "LCMV") {
	    this->bfType = BF::BfLCMV;
	} else if(bfMethod == "GJ") {
	    this->bfType = BF::BfGJ;
	} else if(bfMethod == "GSS") {
	    this->bfType = BF::BfGSS;
	} else if(bfMethod == "GICA") {
	    this->bfType = BF::BfGICA;
	} else if(bfMethod == "GHDSS") {
	    this->bfType = BF::BfGHDSS;
	} else {
	    /* それ以外の値が指定された場合はエラー */
	    throw (new NodeException(NULL,"Unexpected BeamForming:BF_METHOD is set.",__FILE__,__LINE__)) ;
	}

	/* 周波数解析窓長パラメータ値の取得と検証 */
	this->length = dereference_cast<int>(parameters.get("LENGTH"));
	if (this->length <= 0){ 
	    throw (new NodeException(NULL, "BeamForming:LENGTH is invalid value.", __FILE__, __LINE__));
	}

	/* サンプリング周波数パラメータ値の取得と検証 */
	this->samplingRate = dereference_cast<int>(parameters.get("SAMPLING_RATE"));
	if (this->samplingRate <= 0) {
	    throw (new NodeException(NULL, "BeamForming:SAMPLING_RATE is invalid value.", __FILE__, __LINE__));
	}
	
	/* 上下限周波数の取得と検証 */
	int lower_bound_freq = dereference_cast<int>(parameters.get("LOWER_BOUND_FREQUENCY"));
	int upper_bound_freq = dereference_cast<int>(parameters.get("UPPER_BOUND_FREQUENCY"));
	if (lower_bound_freq < 0 || lower_bound_freq > this->samplingRate/2) {
	    throw (new NodeException(NULL, "BeamForming:LOWER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
	}
	if (upper_bound_freq < 0 || upper_bound_freq > this->samplingRate/2) {
	    throw (new NodeException(NULL, "BeamForming:UPPER_BOUND_FREQUENCY is invalid value.", __FILE__, __LINE__));
	}
	if (lower_bound_freq > upper_bound_freq) {
	    throw (new NodeException(NULL, "BeamForming:LOWER_BOUND_FREQUENCY is larger than UPPER_BOUND_FREQUENCY.", __FILE__, __LINE__));
	}
	this->lowerfreq = floor((double)lower_bound_freq / (double)this->samplingRate * (double)this->length);
	this->upperfreq = ceil((double)upper_bound_freq / (double)this->samplingRate * (double)this->length);
	
	/* 分離行列の更新方法とNeighbor判定スレッショルド[m] */
	if (this->bfType == BF::BfLCMV ||
		this->bfType == BF::BfGJ ||
		this->bfType == BF::BfGICA ||
		this->bfType == BF::BfGHDSS) {
 	        
		string update_Method_W = object_cast<string>(parameters.get("UPDATE_METHOD_W"));

		if (update_Method_W == "ID") {
			this->separationMatUpdateType = BF::UtID;
		}
		else if (update_Method_W == "POS") {
			this->separationMatUpdateType = BF::UtPOS;
		}
		else if (update_Method_W == "ID_POS") {
			this->separationMatUpdateType = BF::UtID_POS;
		}
		else {
	    	/* それ以外の値が指定された場合はエラー */
		    throw (new NodeException(NULL,"BeamForming: Unexpected UPDATE_METHOD_W is set.",__FILE__,__LINE__)) ;
		}
	
	    /* Neighbor判定スレッショルド[m] */
      	this->neighborthresh = dereference_cast<float>(parameters.get("UPDATE_ACCEPT_DISTANCE"));
	    this->neighborthresh = this->neighborthresh/1000.0;
	    if (this->neighborthresh < 0.0) {
		    throw (new NodeException(NULL, "BeamForming:UPDATE_ACCEPT_DISTANCE is invalid value.", __FILE__, __LINE__));
		}

	    /* 分離行列初期値（指定されなかった場合はTF_CONJと同じ値となる） */
	    if (parameters.exist("INITW_FILENAME") == true) {
	    	this->initw_filename = object_cast<string>(parameters.get("INITW_FILENAME")) ;
	    	this->iFlgW = 1 ;
	    }
	    else {
	      this->initw_filename = "" ;
	      this->iFlgW = 0 ;
	    }

	    /* 分離行列のエクスポート機能 */
	    bool export_w = dereference_cast<bool>(parameters.get("EXPORT_W"));
	    if (export_w && (parameters.exist("EXPORT_W_FILENAME") == true)) {
	        this->saveW = 1;
	        this->WFileToSave = object_cast<string>(parameters.get("EXPORT_W_FILENAME")) ;
	    }
	    else {
	        this->saveW = 0;
	        this->WFileToSave = "";
	    }
	}

#if 0   
	/* 伝達関数の決定方法 */
	string tf_conj = "DATABASE"; //object_cast<string>(parameters.get("TF_CONJ"));
	if (tf_conj == "DATABASE") {
	    /* 伝達関数の決定方法が DATABASE である場合は、伝達関数の定義ファイルを要求する。 */
#endif
	    if (parameters.exist("TF_CONJ_FILENAME") == true) {
			this->tfFileName = object_cast<string>(parameters.get("TF_CONJ_FILENAME"));
	    } else {
			throw (new NodeException(NULL,"BeamForming:TF_CONJ_FILENAME is not set.",__FILE__,__LINE__)) ;
	    }
#if 0
	    /* 音速値 */
	    this->soundSpeed = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));

	    this->tfType = TfDatabase;

	}
	else if (tf_conj == "CALC") {
	    /* 伝達関数の決定方法が CALC である場合は、マイク位置の定義ファイルを要求する。 */
	    if (parameters.exist("MIC_FILENAME") == true) {
			this->micFileName = object_cast<string>(parameters.get("MIC_FILENAME"));
	    } else {
			throw (new NodeException(NULL,"BeamForming:MIC_FILENAME is not set.",__FILE__,__LINE__)) ;
	    }

	    /* 伝達関数の決定方法が CALC である場合は、マイク座標原点の設定を要求する。 */
	    string micPosShift = object_cast<string>(parameters.get("MIC_POS_SHIFT"));
	    if (micPosShift == "FIX") {
			this->micShiftType = MicShiftFixed;
	    } else if (micPosShift == "SHIFT") {
			this->micShiftType = MicShiftShifted;
	    } else {
			throw (new NodeException(NULL,"BeamForming:MIC_FILENAME is not set.",__FILE__,__LINE__)) ;
	    }
	    
	    /* 音速値 */
	    this->soundSpeed = dereference_cast<float>(parameters.get("SPEED_OF_SOUND"));
		
	    this->tfType = TfCalc;

	}
	else {
	    // それ以外の値が指定された場合はエラー
	    throw (new NodeException(NULL,"Unexpected BeamForming:TF_CONJ is set.",__FILE__,__LINE__)) ;
	}
#endif

	switch (this->bfType) {
#ifdef GSS_ENABLE
	case BF::BfGSS:
		/* 高次相関行列による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GSS_SS_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = false;
		} else if ("LC_MYU" == object_cast<string>(parameters.get("GSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GSS_LC_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = true;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GSS_SS_MYU"));
			this->enableAdaptivessMyu = true;
		}

		/* 高次相関行列計算におけるスケールファクタ */
		if (parameters.exist("GSS_SS_SCAL")){
		    this->ssScalingFactor = dereference_cast<float>(parameters.get("GSS_SS_SCAL"));
		}

		/* 幾何制約による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GSS_LC_METHOD"))){
	  		this->lcMyu = dereference_cast<float>(parameters.get("GSS_LC_MYU"));
			this->enableAdaptivelcMyu = false;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GSS_LC_METHOD"))) {
			this->lcMyu = dereference_cast<float>(parameters.get("GSS_LC_MYU"));
			this->enableAdaptivelcMyu = true;
		}

		break;
#endif
	case BF::BfGICA:
		/* 高次相関行列による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GICA_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GICA_SS_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = false;
		} else if ("LC_MYU" == object_cast<string>(parameters.get("GICA_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GICA_LC_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = true;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GICA_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GICA_SS_MYU"));
			this->enableAdaptivessMyu = true;
		}

		/* 高次相関行列計算におけるスケールファクタ */
		if (parameters.exist("GICA_SS_SCAL")) {
		    this->ssScalingFactor = dereference_cast<float>(parameters.get("GICA_SS_SCAL"));
		}

		/* 幾何制約による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GICA_LC_METHOD"))) {
	  		this->lcMyu = dereference_cast<float>(parameters.get("GICA_LC_MYU"));
			this->enableAdaptivelcMyu = false;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GICA_LC_METHOD"))) {
			this->lcMyu = dereference_cast<float>(parameters.get("GICA_LC_MYU"));
			this->enableAdaptivelcMyu = true;
		}
		break;
	case BF::BfGHDSS:
		/* 高次相関行列による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GHDSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GHDSS_SS_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = false;
		} else if ("LC_MYU" == object_cast<string>(parameters.get("GHDSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GHDSS_LC_MYU"));
			this->enableAdaptivessMyu = false;
			this->useLcMyuAsSsMyu = true;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GHDSS_SS_METHOD"))) {
			this->ssMyu = dereference_cast<float>(parameters.get("GHDSS_SS_MYU"));
			this->enableAdaptivessMyu = true;
			this->useLcMyuAsSsMyu = false;
		}

		/* 高次相関行列計算におけるスケールファクタ */
		if (parameters.exist("GHDSS_SS_SCAL")) {
		    this->ssScalingFactor = dereference_cast<float>(parameters.get("GHDSS_SS_SCAL"));
		}

		/* 幾何制約による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GHDSS_LC_METHOD")) ){
	  		this->lcMyu = dereference_cast<float>(parameters.get("GHDSS_LC_MYU"));
			this->enableAdaptivelcMyu = false;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GHDSS_LC_METHOD"))) {
			this->lcMyu = dereference_cast<float>(parameters.get("GHDSS_LC_MYU"));
			this->enableAdaptivelcMyu = true;
		}

		/* 入力信号をノイズとみなす振幅の閾値(上限) */
		this->ghdss_NoiseFloor = dereference_cast<float>(parameters.get("GHDSS_NOISE_FLOOR"));

		/* 幾何制約の手法 */
		if ("DIAG" == object_cast<string>(parameters.get("GHDSS_LC_CONST"))) {
			this->ghdss_LcConstType = BF::LcDIAG;
		} else if ("FULL" == object_cast<string>(parameters.get("GHDSS_LC_CONST"))) {
			this->ghdss_LcConstType = BF::LcFULL;
		} else {
	    	/* それ以外の値が指定された場合はエラー */
		    throw (new NodeException(NULL,"BeamForming: Unexpected GHDSS_LC_CONST is set.",__FILE__,__LINE__)) ;
		}

		/* 分離行列の更新方法 */
		if ("STEP" == object_cast<string>(parameters.get("GHDSS_UPDATE"))) {
			this->ghdss_UpdateStepType = BF::UsStep;
		} else if ("TOTAL" == object_cast<string>(parameters.get("GHDSS_UPDATE"))) {
			this->ghdss_UpdateStepType = BF::UsTotal;
		} else {
	    	/* それ以外の値が指定された場合はエラー */
		    throw (new NodeException(NULL,"BeamForming: Unexpected GHDSS_UPDATE is set.",__FILE__,__LINE__)) ;
		}
		break;
	case BF::BfLCMV:
		/* 幾何制約による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("LCMV_LC_METHOD"))) {
	  		this->lcMyu = dereference_cast<float>(parameters.get("LCMV_LC_MYU"));
			this->enableAdaptivelcMyu = false;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("LCMV_LC_METHOD"))) {
			this->lcMyu = dereference_cast<float>(parameters.get("LCMV_LC_MYU"));
			this->enableAdaptivelcMyu = true;
		}
		break;
	case BF::BfGJ:
		/* 幾何制約による分離行列更新時のステップサイズ */
		if ("FIX" == object_cast<string>(parameters.get("GJ_LC_METHOD"))) {
	  		this->lcMyu = dereference_cast<float>(parameters.get("GJ_LC_MYU"));
			this->enableAdaptivelcMyu = false;
		} else if ("ADAPTIVE" == object_cast<string>(parameters.get("GJ_LC_METHOD"))) {
			this->lcMyu = dereference_cast<float>(parameters.get("GJ_LC_MYU"));
			this->enableAdaptivelcMyu = true;
		}
		break;
	}
	  	
	/* ICA, HDSS で用いる非線形関数の種類 */
#if 0
	if (parameters.exist("NL_FUNC")) {
	    string nlFunc = object_cast<string>(parameters.get("NL_FUNC"));
	    if (nlFunc == "TANH") {
			/* tanh() の場合*/
#endif
			this->nlFuncType = BF::NlFuncTanh;
#if 0
	    } else {
			/* それ以外の値が指定された場合はエラー。 */
			throw (new NodeException(NULL,"Unexpected BeamForming:NL_FUNC is set.",__FILE__,__LINE__)) ;
	    }
	}
#endif

	/* デバッグ出力用フラグ */
	if (parameters.exist("ENABLE_DEBUG")) {
	  this->enable_debug = dereference_cast<bool>(parameters.get("ENABLE_DEBUG"));
	}
	
	/*== パラメータに基づいて初期化 ==*/

	/* 伝達関数データベース構築 */
#if 0 // 2014/10/16 初期値読込みのためのテスト
	BF::TransformFunctionDatabase newTransformFunctionDatabase;
	if (this->tfType == TfDatabase) {
	    /* HARK 形式のデータベース読み込み */
	    MICARY_TFDBFileHeader tfdbh;
	    CHarkFileIO harkFileIO;
	    if (!harkFileIO.ReadHeader(this->tfFileName.c_str(), &tfdbh)) {
			throw new NodeException(this, "Failed to open TF data file <" + this->tfFileName + ">.", __FILE__, __LINE__);
	    }

	    MICARY_TFDatabase tfdb;
	    TFDatabaseAlloc(&tfdb, tfdbh.lNFreq, tfdbh.lNSrc, tfdbh.lNMic);
	    if (!harkFileIO.ReadInfo(this->tfFileName.c_str(), &tfdbh, &tfdb.micary_locaSource, &tfdb.micary_locaMic, &tfdb.micary_iaFreqLine)
	       || !harkFileIO.ReadData(this->tfFileName.c_str(), &tfdb.micary_cmpaTF)) {
			throw new NodeException(this, "Failed to read TF data file <" + this->tfFileName + ">.", __FILE__, __LINE__);
	    }

	    /* データベース形式を変換 */
	    int numSrc  = tfdb.micary_locaSource.ha_numLocation;
	    int numMic  = tfdb.micary_locaMic.ha_numLocation;
	    int numFreq = tfdb.micary_iaFreqLine.ha_numValue;
	    for (int i = 0; i < numSrc; i++) {
			MICARY_Location &l = tfdb.micary_locaSource.micary_Location[i];
			BF::SourceLocation sl(l.ha_fX, l.ha_fY, l.ha_fZ, tfdb.micary_locaSource.ha_idLocation[i], true);
			newTransformFunctionDatabase.sourceLocations.push_back(sl);
	    }
	    for (int i = 0; i < numMic; i++) {
			MICARY_Location &l = tfdb.micary_locaMic.micary_Location[i];
			BF::MicLocation ml(l.ha_fX, l.ha_fY, l.ha_fZ);
			newTransformFunctionDatabase.micLocations.push_back(ml);
	    }
	    for (int i = 0; i < numFreq; i++) {
			BF::Matrix tf(numMic, numSrc);

			for (int j = 0; j < numMic; j++) {
				for (int k = 0; k < numSrc; k++) {
					HA_Complex &v =tfdb.micary_cmpaTF.ha_cmpValue[(i * numSrc + k) * numMic + j];
					// 2012/05/24 added by N.Tanaka H=conj(TFCONJ)
					//tf.setElementAt(j, k, v.re, v.im);
					tf.setElementAt(j, k, v.re, -v.im);
				}
			}
			newTransformFunctionDatabase.transformFunctions.push_back(tf);
	    }
		TFDatabaseFree(&tfdb);
	}

	if (this->tfType == TfCalc) {
	    /* マイク位置読み込み */
	    MICARY_LocationArray micPositions;
	    CHarkFileIO harkFileIO;
	    LocationArrayAlloc(&micPositions, 0);
	    if (!harkFileIO.LoadPositionData(this->micFileName.c_str(), &micPositions, __FILE_TYPE_MIC)) {
			throw new NodeException(NULL, "Failed to Read mic position file.", __FILE__, __LINE__);
	    }

	    std::vector<BF::MicLocation> newMicPositions;
	    for(int i = 0; i < micPositions.ha_numLocation; i++) {
			const MICARY_Location &mic = micPositions.micary_Location[i];
			newMicPositions.push_back(BF::MicLocation(mic.ha_fX, mic.ha_fY, mic.ha_fZ));
	    }
	    this->micPositions.swap(newMicPositions);

	} else if(this->tfType == TfDatabase) {
	    this->micPositions = newTransformFunctionDatabase.micLocations;
	}
#endif
	CHarkFileIOByMatrix	cHarkMat;
	BF::TransformFunctionDatabase newTransformFunctionDatabase;
	cHarkMat.ReadTFData(this->tfFileName.c_str(), this->neighborthresh, newTransformFunctionDatabase);

	/*== 分離行列初期値の初期化 ==*/

#if 0 // 2014/10/16 初期値読込みのためのテスト
	BF::TransformFunctionDatabase newInitSeparationMatrixDatabase;
	if (this->iFlgW == 0) {
		int numSrc = newTransformFunctionDatabase.sourceLocations.size();
		for (int i = 0; i < numSrc; i++) {
	    	newInitSeparationMatrixDatabase.sourceLocations.push_back(newTransformFunctionDatabase.sourceLocations[i]);
		}
		int numMic = newTransformFunctionDatabase.micLocations.size();
		for (int i = 0; i < numMic; i++) {
			newInitSeparationMatrixDatabase.micLocations.push_back(newTransformFunctionDatabase.micLocations[i]);
		}
		int numFreq = newTransformFunctionDatabase.transformFunctions.size();
		for (int i = 0; i < numFreq; i++) {
			newInitSeparationMatrixDatabase.transformFunctions.push_back(newTransformFunctionDatabase.transformFunctions[i]);
		}
	}
#endif
	BF::Database newSeparationMatricesDatabase;
	if (this->iFlgW == 1) {
		cHarkMat.ReadWData(this->initw_filename.c_str(), newSeparationMatricesDatabase);
	}

	/* adaptBF 作業変数確保 */
	switch(this->bfType){
	case BF::BfDS:    this->bfAdapter = new BF::DS   (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed); break;
	case BF::BfWDS:   this->bfAdapter = new BF::WDS  (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed); break;
	case BF::BfNull:  this->bfAdapter = new BF::Null (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed); break;
	case BF::BfILSE:  this->bfAdapter = new BF::ILSE (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed); break;
	case BF::BfLCMV:  this->bfAdapter = new BF::LCMV (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->enableAdaptivelcMyu, this->lcMyu, this->separationMatUpdateType); break;
	case BF::BfGJ:    this->bfAdapter = new BF::GJ   (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->enableAdaptivelcMyu, this->lcMyu, this->separationMatUpdateType); break;
	case BF::BfGICA:  this->bfAdapter = new BF::GICA (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->enableAdaptivelcMyu, this->lcMyu, this->enableAdaptivessMyu, this->ssMyu, this->ssScalingFactor, this->nlFuncType, this->separationMatUpdateType); break;
#ifdef GSS_ENABLE	
	case BF::BfGSS:   this->bfAdapter = new BF::GSS  (this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->enableAdaptivelcMyu, this->lcMyu, this->enableAdaptivessMyu, this->ssMyu, this->ssScalingFactor, this->nlFuncType, this->separationMatUpdateType); break;	
#endif	
	case BF::BfGHDSS: this->bfAdapter = new BF::GHDSS(this->length / 2 + 1, this->lowerfreq, this->upperfreq, this->samplingRate, this->soundSpeed, this->enableAdaptivelcMyu, this->lcMyu, this->enableAdaptivessMyu, this->ssMyu, this->ssScalingFactor, this->nlFuncType, this->separationMatUpdateType, this->ghdss_NoiseFloor, this->ghdss_LcConstType, this->ghdss_UpdateStepType, this->useLcMyuAsSsMyu); break;
	default:
	    throw new NodeException(NULL, "Adapter not implemented.", __FILE__, __LINE__);
	}

	/* 伝達関数データベースを登録 */
#if 0
	if (this->tfType == TfDatabase) {
#endif
	    this->bfAdapter->setTransformFunctionDatabase(newTransformFunctionDatabase);
#if 0
	} else {
	    this->bfAdapter->newTransformFunctionDatabase();
	}
#endif

	/* 分離行列データベースの初期値を登録 */
	if (this->iFlgW == 1) {
	    this->bfAdapter->setSeparationMatricesDatabase(newSeparationMatricesDatabase);
	} else { 
	    this->bfAdapter->newSeparationMatricesDatabase();
	}
	}

    virtual ~BeamForming()
    {
    /*== 分離行列データベースの最終保存 ==*/
    this->bfAdapter->updateSeparationMatricesDatabaseAll();

    /*== 分離行列のエクスポート ==*/
    if (this->saveW == 1) {
		this->exportSeparationMatrices(this->WFileToSave, this->samplingRate, this->length, this->bfAdapter->separationMatricesDatabase);
    }

	delete this->bfAdapter;
    }

public:
    virtual int
    translateInput(const string &inputName)
    {
	if (inputName == "INPUT_FRAMES") {
		return this->input_frameID;
	} else if(inputName == "INPUT_SOURCES") {
		return this->input_signalSourceID;
	} else if(inputName == "INPUT_NOISE_SOURCES") {
		return this->input_noiseSourceID = addInput("INPUT_NOISE_SOURCES");
	} else {
		throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
	}
    }

    void calculate(int output_id, int count, Buffer &out)
    {
//std::cout << "BF->" << count << std::endl;
	/*== 入出力の準備 ==*/

	/* スペクトル入力 */
	ObjectRef orInputTemp = getInput(this->input_frameID, count);
	Matrix<complex<float> > &inputValueMatrix = object_cast<Matrix<complex<float> > >(orInputTemp);
	
	int numMic  = inputValueMatrix.nrows();	// チャンネル数
	int numFreq = inputValueMatrix.ncols();	// 周波数ライン数
	    	    
	/* 信号音源位置情報入力 */
	RCPtr<Vector<ObjectRef> > pSignalSources = getInput(this->input_signalSourceID, count);
	const Vector<ObjectRef> *signalSources = pSignalSources.get();
	int numSourceS = signalSources->size();	 // 信号音源数
	    
	/* ノイズ音源位置情報入力 */
	RCPtr<Vector<ObjectRef> > pNoiseSources;
	const Vector<ObjectRef> *noiseSources = NULL;
	int numSourceN = 0;
	if (this->input_noiseSourceID >= 0) {
	    pNoiseSources = getInput(this->input_noiseSourceID, count);
	    noiseSources = pNoiseSources.get();
	    numSourceN = noiseSources->size(); // ノイズ音源数
	}

	int numSource  = numSourceS + numSourceN; // 総音源数

	/* 出力スペクトルの領域を準備 */
	ObjectRef orOutputTemp = ObjectRef(new Map<int,ObjectRef>);
	Map<int, ObjectRef> &mpOutputValue = object_cast<Map<int,ObjectRef> >(orOutputTemp);
	out[count] = orOutputTemp;
	mpOutputValue.clear();

	/*== 処理用データの準備 ==*/

	/* 入力スペクトル */
	BF::Matrix inputSpectrum(numMic, numFreq);
	for (int i = 0 ; i < numMic; i++) {
	    for (int j = 0; j < numFreq; j++) {
			complex<float> v = inputValueMatrix(i, j);
			inputSpectrum.setElementAt(i, j, BF::Complex(v.real(), v.imag()));
	    }
	}

	/* 出力スペクトル */
	BF::Matrix outputSpectrum(numSource, numFreq);

	/* 音源位置 */
	std::vector<BF::SourceLocation> sourcePositions;
	for (int i = 0 ; i < numSource; i++) {
	    RCPtr<Source> src = (i < numSourceS) ? signalSources->at(i) : noiseSources->at(i - numSourceS);
	    if (src->compare_mode == Source::TFINDEX) {
	        sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], src->tfindex[0], src->id, (i < numSourceS)));
	    } else { // Source::DEG
	        sourcePositions.push_back(BF::SourceLocation(src->x[0], src->x[1], src->x[2], -1, src->id, (i < numSourceS)));
	        int idx = this->bfAdapter->transformFunctionDatabase->findNearestSourceIndex(sourcePositions[sourcePositions.size()-1]);
	        sourcePositions[sourcePositions.size()-1].id = this->bfAdapter->transformFunctionDatabase->sourceLocations[idx].id;
	    }
	}

	if (numSource > 0) {
	    /*== 伝達関数適用（更新または継承）および分離行列初期化 ==*/
	    this->bfAdapter->updateSources(sourcePositions);

	    /*== BeamForming 実行 ==*/
	    this->bfAdapter->executeBeamForming(inputSpectrum, outputSpectrum);

	    /*== 分離行列適応処理 ==*/
	    this->bfAdapter->adaptSeparationMatrices();

	    /*== POS, ID_POSのとき、分離行列データベースへの登録 ==*/
	    if (this->bfAdapter->separationMatUpdateType == BF::UtPOS || this->bfAdapter->separationMatUpdateType == BF::UtID_POS) {
		    this->bfAdapter->updateSeparationMatricesDatabaseAll();
	    }

	    /*== 分離行列のエクスポート ==*/
	    /*if (this->saveW) {
			this->exportSeparationMatrices(this->WFileToSave, this->length / 2 + 1, this->samplingRate, this->micPositions, sourcePositions, this->bfAdapter->separationMatrices);
	    }*/	

	    /*== 処理結果を出力スペクトルの領域へ格納 ==*/
	    for (int i = 0 ; i < numSource; i++) {
			int id = sourcePositions[i].sourceID;
			ObjectRef orOutputTempV = ObjectRef(Vector<complex<float> >::alloc(numFreq));
			Vector<complex<float> > &vcfOutputValue = object_cast<Vector<complex<float> > > (orOutputTempV) ;
		    
			for (int j = 0 ; j < numFreq; j++) {
#ifdef USING_HARKIO3_MATRIX
		    	BF::Complex v = outputSpectrum.getElementAt(i, j);			
#else
		    	const BF::Complex v = outputSpectrum.getElementAt(i, j);			
#endif
				float re = v.real();
				float im = v.imag();
				if ((fpclassify(re) == FP_NAN || fpclassify(im) == FP_NAN) && (this->enable_debug)) {
					std::cout << "BeamForming: Output Spectrum data ["<< i <<" , " <<  j << "] is NAN. Please check the parameters." << std::endl ;
				}
		    	if ((fpclassify(re) == FP_INFINITE || fpclassify(im) == FP_INFINITE) && (this->enable_debug)) {
					std::cout << "BeamForming: Output Spectrum data ["<< i <<" , " <<  j << "] is INF. Please check the parameters." << std::endl ;
		    	}
		    	vcfOutputValue[j] = complex<float>(re, im);
			}
			mpOutputValue.insert(make_pair(id,orOutputTempV));
	    }
	}
    }

#if 0 // HarkFileIOMatrixを使用するため
    void
    exportSeparationMatrices(const string              &filename,          /* 出力先ファイル名 */
			     int                                   numFreq,            /* 周波数ビンの数 */
			     int                                   samplingRate,       /* サンプリング周波数 */
			     const std::vector<BF::MicLocation>    &micPositions,      /* マイク位置 */
			     const std::vector<BF::SourceLocation> &sourcePositions,   /* 音源位置 */
			     const std::vector<BF::Matrix>         &separationMatrices /* 最新の分離行列群 */)
    {
	int numMic = micPositions.size();
	int numSrc = sourcePositions.size();

	MICARY_SMDatabase smDatabase = {0};
		
	SMDatabaseAlloc(&smDatabase, 0, 0, 0);
	SMDatabaseAlloc(&smDatabase, numFreq, numSrc, numMic);

	for (int i = 0; i < numMic; i++) {
	    const BF::MicLocation &ml = micPositions[i];
	    smDatabase.micary_locaMic.ha_idLocation[i] = i;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fX = ml.x;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fY = ml.y;
	    smDatabase.micary_locaMic.micary_Location[i].ha_fZ = ml.z;
	    smDatabase.micary_locaMic.micary_Location[i].ha_idPos = i;
	}

	for (int i = 0; i < numSrc; i++) {
	    const BF::SourceLocation &sl = sourcePositions[i];
	    smDatabase.micary_locaSource.ha_idLocation[i] = sl.sourceID;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fX = sl.x;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fY = sl.y;
	    smDatabase.micary_locaSource.micary_Location[i].ha_fZ = sl.z;
	    smDatabase.micary_locaSource.micary_Location[i].ha_idPos = sl.sourceID;
	}

	for (int i = 0 ; i < numFreq; i++) {
	    smDatabase.micary_smSeparation.micary_iaFreqLinex.ha_iValue[i] = (double)samplingRate * (i * 2 + 1) / 4 / numFreq;
	}

	int c = 0;
	for (int i = 0 ; i < numFreq; i++) {
	    const BF::Matrix &mat = separationMatrices[i];

	    for (int j = 0 ; j < numSrc; j++) {
			for (int k = 0 ; k < numMic; k++) {
			    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].re = mat.getElementAt(j, k).real();
			    smDatabase.micary_smSeparation.micary_cmpaSeparation.ha_cmpValue[c].im = mat.getElementAt(j, k).imag();
			    c++;
			}
	    }
	}

	MICARY_TFDBFileHeader tfdbfh;
	tfdbfh.lFs       = samplingRate;
	tfdbfh.lNFFT     = numFreq * 2;
	tfdbfh.lNMic     = numMic;
	tfdbfh.lNSrc     = numSrc;
	tfdbfh.lNFreq    = numFreq;
	tfdbfh.iVersion  = 1;
	tfdbfh.iFileType = FORMAT_TYPE_W;

	CHarkFileIO cHark;
	if (!cHark.WriteData(filename.c_str(), &tfdbfh,
			    smDatabase.micary_smSeparation.micary_cmpaSeparation,
			    smDatabase.micary_smSeparation.micary_iaFreqLinex,
			    smDatabase.micary_locaSource,
			    smDatabase.micary_locaMic,
			    "",
			    "",
			    "",
			    "",
			    "",
			    "GHDSS_EXPORT")) {
	    SMDatabaseFree(&smDatabase);
	    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
	}

	SMDatabaseFree(&smDatabase);
    }
#endif

    void
    exportSeparationMatrices(const string  &filename, /* 出力先ファイル名 */
							int samplingRate,         /* サンプリング周波数 */
							int numFreq,              /* 周波数ビンの数 */
    						BF::Database *SeparationMatricesDatabase)
    {
	CHarkFileIOByMatrix	cHarkMat;
	if (!SeparationMatricesDatabase) {
		return;
	}
	if (!cHarkMat.WriteWData(filename.c_str(), samplingRate, numFreq, *(SeparationMatricesDatabase))) {
	    throw new NodeException(this, "Fail to Export file as HARK format.", __FILE__, __LINE__);
	}
	}


private:
    /*=== FlowDesigner 入出力 ===*/

    /* フレームデータの入力ID */
    int input_frameID;

    /* 信号音源情報の入力ID */
    int input_signalSourceID;

    /* ノイズ音源情報の入力ID */
    int input_noiseSourceID;

    /* 処理結果フレームデータの出力ID */
    int output_ID;

    /*=== パラメータ ===*/

    /* BF アルゴリズム種類 */
    BF::BfType bfType;

#if 0
    /* 伝達関数をファイルから読み込んで使用するか？ */
    enum TfType {
	TfUnknown = 0, TfDatabase, TfCalc,
    };
    TfType tfType;
#endif

    /* 伝達関数をファイルから読み込んで使用する場合のファイル名 */
    string tfFileName;

#if 0
    /* マイク原点位置の設定 */
    enum MicShiftType {
	MicShiftUnknown = 0, MicShiftFixed, MicShiftShifted,
    };
    MicShiftType micShiftType;

    /* マイク位置の定義ファイル名 */
    string micFileName;
#endif

    /* 分離行列をファイルへ書き込むか？書き込む場合はそのファイル名 */
    int    saveW;
    string WFileToSave;

    /* 周波数解析窓長 */
    int length;

    /* サンプリング周波数 */
    int samplingRate;

    /* 音速値 */
    float soundSpeed;

    /* 分析対象周波数 */
    int lowerfreq;
    int upperfreq;
    
    /* 分離行列の更新方法 */
	BF::UtUpdateType separationMatUpdateType;
	
    /* Neighbor判定スレッショルド[m] */
    float neighborthresh;

    /* 分離行列初期値ファイル */
    int iFlgW ;
    string initw_filename;

    /* 高次相関行列計算におけるスケールファクタ */
    float ssScalingFactor;

    /* 高次相関行列による分離行列更新時のステップサイズ */
    bool enableAdaptivessMyu;
    bool useLcMyuAsSsMyu;
    float ssMyu;

    /* 幾何制約による分離行列更新時のステップサイズ */
    bool enableAdaptivelcMyu;
    float lcMyu;

    /* ICA, HDSS で用いる非線形関数の種類 */
    BF::NlFuncType nlFuncType;
	
    /* (GHDSS) 入力信号をノイズとみなす振幅の閾値(上限)  */
    float ghdss_NoiseFloor;
    
    /* (GHDSS) 幾何制約の手法 */
    BF::LcConstType ghdss_LcConstType;
    
    /* (GHDSS) 分離行列の更新方法 */
    BF::UsUpdateStepType ghdss_UpdateStepType;
    
    /* デバッグ出力用フラグ */
    bool enable_debug;

    /*=== 作業用変数 ===*/

#if 0
    /* マイク位置 (tfType == TfCalc 時のみ) */
    std::vector<BF::MicLocation> micPositions;
#endif

    /* 分離行列の更新を行う処理オブジェクト  */
    BF::BF *bfAdapter;

};

